/* GTKabale - Solitaire game for lazy people

   Copyright (C) 1999 Ren Seindal (rene@seindal.dk)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
 */

#include <stdlib.h>
#include <time.h>
#include "data.h"

#include <glib.h>

struct card_t *deck = nil;
struct card_t *top = nil;

stack_t layout[8];
value_t aces[4];

int cardsleft = 0, turnsleft = 0;
int times_played = 0, times_won = 0;


typedef unsigned char deckindex_t;

static card_t deckdata[CARDS];


static void
swapdeck(deckindex_t *i, deckindex_t *j)
{
    deckindex_t x;

    x = *i; 
    *i = *j;
    *j = x;
}

void
initdata(gboolean same)
{
    suit_t s;
    value_t v;
    stacknum_t sn;
    deckindex_t d[CARDS];
    deckindex_t i, j;
    static guint seed = 0;

    if (!same)
	seed = (guint)time(nil);
    srand(seed);

    for (s = hearts; s <= clubs; s++) 
	for (v = ace; v <= king; v++) {
	    i = s * 13 + v - 1;
	    deckdata[i].s = s;
	    deckdata[i].v = v;
	    deckdata[i].n = nil;
	}

    for (i = 0; i < CARDS; i++)
	d[i] = i;

    for (i = CARDS; --i > 0; ) {
	j = rand() % (i+1);
	swapdeck(d+i, d+j);
    }

    i = 0; /* takeoff index for layout */
    for (sn = 1; sn <= STACKS; sn++) {
	layout[sn].hidden = nil;
	j = 1;
	while (j < sn) {
	     deckdata[d[i]].n = layout[sn].hidden;
	     layout[sn].hidden = &deckdata[d[i]];
	     i++;
	     j++;
	};
	layout[sn].faceup = &deckdata[d[i]];
	layout[sn].last = layout[sn].faceup;
	i++;
    }

    top = nil;
    for ( ; i < CARDS; i++) {
	deckdata[d[i]].n = top;
	top = &deckdata[d[i]];
    }

    deck = nil;
    for (s = hearts; s <= clubs; s++)
	aces[s] = none;

    cardsleft = CARDS;
    turnsleft = 3;
}

boolean
succes(void)
{
    return (cardsleft == 0);
}

boolean
failure(void)
{
    return (turnsleft <= 0);
}

boolean
same_color(card_t *s, card_t *d)
{
    return
	(((d->s == diamonds) || (d->s == hearts)) &&
	 ((s->s == diamonds) || (s->s == hearts))) ||
	(((d->s == spades) || (d->s == clubs)) &&
	 ((s->s == spades) || (s->s == clubs)));
}

boolean
can_move_stack(card_t *s, card_t *d)
{
    return
	(s != nil) &&
	(((d == nil) &&
	  (s->v == king)) ||
	 ((d != nil) &&
	  (!same_color(s, d)) &&
	  (s->v == pred(d->v)))
	 );
}

boolean
can_move_ace(card_t *s)
{
    return (s != nil) && (s->v == succ(aces[s->s]));
}

card_t *
getlast(card_t *list, card_t **prev)
{
   card_t *last;

   last = list;
   *prev = nil;

   if (last != nil) {
       while (last->n != nil) {
	   *prev = last;
	   last = last->n;
       }
   }
   return last;
}

void
flip_card(stacknum_t sn)
{
    card_t *last, *prev;

    last = getlast(layout[sn].hidden, &prev);
    layout[sn].faceup = last;
    if (last != nil) {
	if (prev == nil)
	    layout[sn].hidden = nil;
	else
	    prev->n = nil;
    }
    layout[sn].last = last;
}

boolean
move_stack_to_ace(stacknum_t src)
{
    card_t *last, *prev;
 
    last = getlast(layout[src].faceup, &prev);

    if (can_move_ace(last)) {
	if (prev == nil)
	    flip_card(src);
	else {
	    prev->n = nil;
	    layout[src].last = prev;
	}
	aces[last->s]++;
	--cardsleft;
	return TRUE;
    } else
	return FALSE;
}

boolean
move_stacks(stacknum_t src, stacknum_t dest)
{
    card_t *s, *d;

    s = layout[src].faceup;
    d = layout[dest].last;

    if (can_move_stack(s, d)) {
	if (d == nil) 
	    layout[dest].faceup = s;
	else
	    d->n = s;
	layout[dest].last = layout[src].last;
	flip_card(src);
	return TRUE;
    } else
	return FALSE;
}

boolean
move_deck_to_stack(stacknum_t dest)
{
    card_t *d;

    d = layout[dest].last;

    if (can_move_stack(deck, d)) {
	if (d == nil)
	     layout[dest].faceup = deck;
	else
	     d->n = deck;

	layout[dest].last = deck;

	d = deck;
	deck = deck->n;
	d->n = nil;
	return TRUE;
    } else
	return FALSE;
}

boolean
move_deck_to_ace()
{
    if (can_move_ace(deck)) {
	aces[deck->s]++;
	--cardsleft;

	deck = deck->n;
	return TRUE;
    } else
	return FALSE;
}

boolean
pop_deck(void)
{
    int i;
    card_t *t;
    boolean retval = FALSE;

    i = 3;
    while ((i > 0) && (top != nil)) {
	--i;
	t = top;
	top = top->n;
	t->n = deck;
	deck = t;
	retval = TRUE;
    }
    return retval;
}

card_t *
revert_deck(card_t *from)
{
    card_t *tmp, *top;

    top = nil;
    while (from != nil) {
	tmp = from;
	from = from->n;
	tmp->n = top;
	top = tmp;
    }
    return top;
}

boolean
null_move(stacknum_t sn)
{
    return (layout[sn].faceup->v == king) 
	&& (layout[sn].hidden == nil);
}

boolean
find_move(stackindex_t *srcp, stackindex_t *destp)
{
    stackindex_t src;
    stackindex_t dest;
    stacknum_t sn, sn2;
    boolean ok;

    ok = FALSE;

    src = 0;
    dest = 0;

    /* check DECK to ACE */
    if (can_move_ace(deck))
	ok = TRUE;

    /* check STACK to ACE */
    if (!ok) {
	for (sn = 1; sn <= STACKS; sn++)
	    if (can_move_ace(layout[sn].last)) {
		src = sn;
		ok = TRUE;
	    }
    }

    /* check STACK to STACK */
    if (!ok) {
	for (sn = 1; sn <= STACKS; sn++)
	    for (sn2 = 1; sn2 <= STACKS; sn2++)
		if (can_move_stack(layout[sn].faceup, layout[sn2].last) &&
		   (!null_move(sn)))
		{
		    src = sn;
		    dest = sn2;
		    ok = TRUE;
		}
    }

    /* check DECK to STACK */
    if (!ok) {
	for (sn = 1; sn <= STACKS; sn++)
	    if (can_move_stack(deck, layout[sn].last)) {
		dest = sn;
		ok = TRUE;
	    }
    }

    if (srcp != NULL)
	*srcp = src;
    if (destp != NULL)
	*destp = dest;

    return ok;
}

int
find_unique_move(stackindex_t src, stackindex_t *dest)
{
    stacknum_t sn;
    int movecnt;

    movecnt = 0;

    if (src == 0) {
	if (deck != nil) {
	    /* check DECK to STACK */
	    if (deck->v != ace) {
		for (sn = 1; sn <= STACKS; sn++)
		    if (can_move_stack(deck, layout[sn].last)) {
			movecnt++;
			*dest = sn;
		    }
	    }

	    /* special check for kings and twos*/
	    if ((deck->v == king) && (movecnt > 1))
		movecnt = 1;
	    if ((deck->v == two) && can_move_ace(deck)) 
		movecnt = 0;

	    /* check DECK to ACE */
	    if (can_move_ace(deck)) {
		movecnt++;
		*dest = 0;
	    }
	}
    } else if (src != 8) {
	if (layout[src].faceup != nil) {
	    /* check STACK to STACK */
	    if (layout[src].last->v != ace) {
		for (sn = 1; sn <= STACKS; sn++)
		    if (can_move_stack(layout[src].faceup, layout[sn].last) &&
		       (!null_move(src))) {
			*dest = sn;
			movecnt++;
		    }
	    }

	    /* special check for kings and twos */
	    if ((layout[src].faceup->v == king) && (movecnt > 1)) 
		movecnt = 1;

	    if ((layout[src].last->v == two) &&
		can_move_ace(layout[src].last)) 
		movecnt = 0;

	    /* check STACK to ACE */
	    if (can_move_ace(layout[src].last)) {
		movecnt++;
		*dest = 0;
	    }
	}
    }

    return movecnt;
}
