/* GTKabale - Solitaire game for lazy people

   Copyright (C) 1999 Ren Seindal (rene@seindal.dk)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#ifdef DEBUG
#define IMAGEDIR PACKAGE_SOURCE_DIR "/cards"
#else
#define IMAGEDIR PACKAGE_DATA_DIR
#endif /* DEFINE */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include <gtk/gtk.h>
#include <gdk_imlib.h>

#include "interface.h"
#include "support.h"

#include "data.h"
#include "display.h"

typedef struct cardinfo_t {
    GdkPixmap *face;
    GdkPixmap *mask;
    GtkWidget *w;
    GdkPixmap *smface;
    GdkPixmap *smmask;
    GtkWidget *sw;
    gboolean is_small;
    gint stack;
    gint ypos;
    enum { CARD_UNSEEN, CARD_VISIBLE, CARD_HIDDEN } state;
} cardinfo_t;

const int card_width = 80;
const int card_height = 120;

const int card_overlap_fraction = 4;
const int card_overlap = 120/5;

const int extra = 3;

static guint layout_width;
static guint layout_height;

static cardinfo_t info[CARDS];
static cardinfo_t back;

static int naces;
static boolean reveal_mode;

static GtkWidget *window;
static GtkFixed *stack[9];

static GtkLabel	*msg_passes;
static GtkLabel	*msg_deck;
static GtkLabel	*msg_rest;
static GtkLabel	*msg_aces;
static GtkLabel	*msg_games;
static GtkLabel	*msg_won;

static GtkWidget *button_new;
static GtkWidget *button_same;

static GtkPixmap *feedback;

struct feedback_images {
    GdkPixmap *face;
    GdkPixmap *mask;
};

static struct feedback_images feedback_images[DISPLAY_NONE];

void
update_screen(void)
{
    while (gtk_events_pending())
	gtk_main_iteration();
}

static void
update_screen_delay(void)
{
    usleep(50000);
    update_screen();
}

static cardinfo_t *
get_info(card_t *c)
{
    if (c->s < 0 || c->s > 3 || c->v < 1 || c->v > 13)
	g_error("bad get_info()");
    return &info[c->s * 13 + c->v - 1];
}

static int
listlength(card_t *c)
{
    int l = 0;
    while (c != nil) {
	l++;
	c = c->n;
    }
    return l;
}

static void
move_card(cardinfo_t *c, stacknum_t sn, guint pos)
{
    if (c->stack == sn)
	return;

#ifdef DEBUG
    g_print("move_card %d/%d %d->%d\n", (c-info)/13, (c-info)%13+1,
	    c->stack, sn);
#endif /* DEBUG */

    gtk_widget_hide(c->w);
    gtk_container_remove(GTK_CONTAINER(stack[c->stack]), c->w);
    gtk_fixed_put(stack[sn], c->w, 0, pos);

    gtk_widget_hide(c->sw);
    gtk_container_remove(GTK_CONTAINER(stack[c->stack]), c->sw);
    gtk_fixed_put(stack[sn], c->sw, 0, pos);

    c->stack = sn;
}

static void
hide_card(cardinfo_t *c)
{
#ifdef DEBUG
    g_print("hide_card %d/%d\n", (c-info)/13, (c-info)%13+1);
#endif /* DEBUG */

    gtk_widget_hide(c->w);
    gtk_widget_hide(c->sw);
}

static void
show_card_small(cardinfo_t *c)
{
#ifdef DEBUG
    g_print("show_card_small %d/%d\n", (c-info)/13, (c-info)%13+1);
#endif /* DEBUG */

    gtk_widget_hide(c->w);
    gtk_widget_show(c->sw);
}

static void
show_card(cardinfo_t *c)
{
#ifdef DEBUG
    g_print("show_card %d/%d\n", (c-info)/13, (c-info)%13+1);
#endif /* DEBUG */

    gtk_widget_show(c->w);
    gtk_widget_hide(c->sw);
}

static void
set_card_hidden(cardinfo_t *c)
{
#ifdef DEBUG
    g_print("set_card_hidden %d/%d\n", (c-info)/13, (c-info)%13+1);
#endif /* DEBUG */

    gtk_pixmap_set(GTK_PIXMAP(c->w), back.face, back.mask);
    gtk_pixmap_set(GTK_PIXMAP(c->sw), back.smface, back.smmask);
}

static void
set_card_visible(cardinfo_t *c)
{
#ifdef DEBUG
    g_print("set_card_visible %d/%d\n", (c-info)/13, (c-info)%13+1);
#endif /* DEBUG */

    gtk_pixmap_set(GTK_PIXMAP(c->w), c->face, c->mask);
    gtk_pixmap_set(GTK_PIXMAP(c->sw), c->smface, c->smmask);
}

static gboolean
image_load_with_size(gchar *file, 
		     GdkPixmap **p, GdkPixmap **m, 
		     guint size_x, guint size_y)
{
    GdkImlibImage *im;

    im = gdk_imlib_load_image(file);
    if (im == NULL)
	g_error("cannot read cards data from %s\n", file);

    if (size_x > 0)
	size_x = im->rgb_width;
    if (size_y > 0)
	size_y = im->rgb_height;

    gdk_imlib_render(im, size_x, size_y);
    if (p)
	*p = gdk_imlib_move_image(im);
    if (m)
	*m = gdk_imlib_move_mask(im);

    gdk_imlib_destroy_image(im);
    return TRUE;
}


static void
init_card(struct cardinfo_t *p)
{
    p->w = gtk_pixmap_new(p->face, p->mask);
    gtk_widget_ref(p->w);

    p->sw = gtk_pixmap_new(p->smface, p->smmask);
    gtk_widget_ref(p->sw);

    p->state = CARD_UNSEEN;
    p->stack = 0;
    p->ypos = 0;
    p->is_small = FALSE;
}

static void
init_graphics(void)
{
    GdkImlibImage *all_cards;
    GdkImlibImage *card;
    guint i;
    guint cw, ch;
    guint s, v;

    gfloat total_count = CARDS + 7;
    guint current_count = 0;

    GtkWidget *progress = create_load_monitor();
    GtkProgressBar *pbar = GTK_PROGRESS_BAR(lookup_widget(progress,
							  "progressbar"));
    gtk_label_set(GTK_LABEL(lookup_widget(progress, "title")),
		  "GTKabale " VERSION);
    gtk_widget_show(progress);

    gtk_progress_bar_update(pbar, current_count++/total_count);
    update_screen();

    all_cards = gdk_imlib_load_image(IMAGEDIR "/cards.png");
    if (all_cards == NULL)
	g_error("cannot read cards data from %s\n", IMAGEDIR "/cards.png");

    gtk_progress_bar_update(pbar, current_count++/total_count);
    update_screen();

    cw = all_cards->rgb_width/13;
    ch = all_cards->rgb_height/4;

    for (i = 0; i < CARDS; i++) {
	s = i/13;
	v = i%13;

	card = gdk_imlib_crop_and_clone_image(all_cards, v*cw, s*ch, cw, ch);
	gdk_imlib_render(card, card_width, card_height);
	info[i].face = gdk_imlib_move_image(card);
	info[i].mask = gdk_imlib_move_mask(card);
	gdk_imlib_destroy_image(card);

	card = gdk_imlib_crop_and_clone_image(all_cards, v*cw, s*ch,
					      cw, card_overlap+extra);
	gdk_imlib_render(card, card_width, card_overlap+extra);
	info[i].smface = gdk_imlib_move_image(card);
	info[i].smmask = gdk_imlib_move_mask(card);
	gdk_imlib_destroy_image(card);

	init_card(&info[i]);

	gtk_fixed_put(stack[0], info[i].w, 0, layout_height-card_height);
	gtk_fixed_put(stack[0], info[i].sw, 0, layout_height-card_height);

	gtk_progress_bar_update(pbar, current_count++/total_count);
	update_screen();
    }
    gdk_imlib_destroy_image(all_cards);

    card = gdk_imlib_load_image(IMAGEDIR "/back.png");
    gdk_imlib_render(card, card_width, card_height);
    back.face = gdk_imlib_move_image(card);
    back.mask = gdk_imlib_move_mask(card);

    gdk_imlib_crop_image(card, 0, 0, card->rgb_width, card_overlap+extra);
    gdk_imlib_render(card, card_width, card_overlap+extra);
    back.smface = gdk_imlib_move_image(card);
    back.smmask = gdk_imlib_move_mask(card);
    gdk_imlib_destroy_image(card);

    init_card(&back);

    gtk_fixed_put(stack[0], back.w, 0, layout_height-2*card_height);
    gtk_fixed_put(stack[0], back.sw, 0, layout_height-2*card_height);

    gtk_progress_bar_update(pbar, current_count++/total_count);
    update_screen();

    image_load_with_size(IMAGEDIR "/no-move.png",
			 &feedback_images[DISPLAY_BAD_MOVE].face, 
			 &feedback_images[DISPLAY_BAD_MOVE].mask,
			 card_width, card_height);

    gtk_progress_bar_update(pbar, current_count++/total_count);
    update_screen();

    image_load_with_size(IMAGEDIR "/help-no.png",
			 &feedback_images[DISPLAY_NO_MOVE].face, 
			 &feedback_images[DISPLAY_NO_MOVE].mask,
			 card_width, card_height);

    gtk_progress_bar_update(pbar, current_count++/total_count);
    update_screen();

    image_load_with_size(IMAGEDIR "/help-yes.png",
			 &feedback_images[DISPLAY_CAN_MOVE].face, 
			 &feedback_images[DISPLAY_CAN_MOVE].mask,
			 card_width, card_height);

    gtk_progress_bar_update(pbar, current_count++/total_count);
    update_screen();

    image_load_with_size(IMAGEDIR "/ambiguous.png",
			 &feedback_images[DISPLAY_AMBIGUOUS_MOVE].face, 
			 &feedback_images[DISPLAY_AMBIGUOUS_MOVE].mask,
			 card_width, card_height);

    gtk_progress_bar_update(pbar, current_count++/total_count);
    update_screen();

    image_load_with_size(IMAGEDIR "/game-won.png",
			 &feedback_images[DISPLAY_WON].face, 
			 &feedback_images[DISPLAY_WON].mask,
			 card_width, card_height);

    gtk_progress_bar_update(pbar, current_count++/total_count);
    update_screen();

    image_load_with_size(IMAGEDIR "/game-lost.png",
			 &feedback_images[DISPLAY_LOST].face, 
			 &feedback_images[DISPLAY_LOST].mask,
			 card_width, card_height);

    gtk_progress_bar_update(pbar, 1.0);
    update_screen();

    gtk_widget_destroy(progress);
}


void
init_display(int argc, char **argv)
{
    gint i;

#ifdef DEBUG
    g_print("init_display\n");
#endif /* DEBUG */

#ifdef DEBUG
    gtk_rc_add_default_file(PACKAGE_SOURCE_DIR "/gtkrc");
#else
    gtk_rc_add_default_file(PACKAGE_DATA_DIR "/gtkrc");
#endif /* DEBUG */

    gtk_set_locale();
    gtk_init(&argc, &argv);
    gdk_imlib_init();

    window = create_GTKabale();
    gtk_widget_set_name(window, "GTKabale");
    gtk_window_set_title(GTK_WINDOW(window), "GTKabale " VERSION);

    stack[0] = GTK_FIXED(lookup_widget(window, "stack0"));
    stack[1] = GTK_FIXED(lookup_widget(window, "stack1"));
    stack[2] = GTK_FIXED(lookup_widget(window, "stack2"));
    stack[3] = GTK_FIXED(lookup_widget(window, "stack3"));
    stack[4] = GTK_FIXED(lookup_widget(window, "stack4"));
    stack[5] = GTK_FIXED(lookup_widget(window, "stack5"));
    stack[6] = GTK_FIXED(lookup_widget(window, "stack6"));
    stack[7] = GTK_FIXED(lookup_widget(window, "stack7"));
    stack[8] = GTK_FIXED(lookup_widget(window, "stack8"));

    msg_passes = GTK_LABEL(lookup_widget(window, "passes"));
    msg_deck = GTK_LABEL(lookup_widget(window, "deck"));
    msg_rest = GTK_LABEL(lookup_widget(window, "rest"));
    msg_aces = GTK_LABEL(lookup_widget(window, "aces"));
    msg_games = GTK_LABEL(lookup_widget(window, "games"));
    msg_won = GTK_LABEL(lookup_widget(window, "won"));

    button_new = lookup_widget(window, "start");
    button_same = lookup_widget(window, "same");

    feedback = GTK_PIXMAP(lookup_widget(window, "feedback"));

    layout_width = 9*card_width;
    layout_height = MAX((17*card_overlap + card_height), (4*card_height+9));

    for (i = 9; --i >= 0; )
	gtk_widget_set_usize(GTK_WIDGET(stack[i]), card_width, layout_height);

    init_graphics();

    gtk_widget_show(window);
}

void
display_clear(void)
{
    gint i;

#ifdef DEBUG
    g_print("display_clear\n");
#endif /* DEBUG */

    for (i = 0; i < CARDS; i++) {
	if (info[i].stack > 0) {
	    move_card(&info[i], 0, layout_height-card_height);
	    hide_card(&info[i]);

	    info[i].state = CARD_UNSEEN;
	}
    }

    naces = 0;
}

void
display_info(void)
{
    gchar buf[256];

#ifdef DEBUG
    g_print("display_info\n");
#endif /* DEBUG */

    sprintf(buf, _("Passes: %c"), turnsleft == 0 ? '-' : 4-turnsleft+'0');
    gtk_label_set(msg_passes, buf);

    sprintf(buf, _("Deck: %2d"), listlength(top)+listlength(deck));
    gtk_label_set(msg_deck, buf);

    sprintf(buf, _("Remain: %2d"), listlength(top));
    gtk_label_set(msg_rest, buf);

    sprintf(buf, _("Aces: %2d"), CARDS - cardsleft);
    gtk_label_set(msg_aces, buf);

    sprintf(buf, _("Games: %2d"), times_played);
    gtk_label_set(msg_games, buf);

    sprintf(buf, _("Won: %2d"), times_won);
    gtk_label_set(msg_won, buf);
}

static gboolean
display_card(card_t *c, boolean hidden, stacknum_t sn, int ypos)
{
    cardinfo_t *cinfo;
    gboolean ret = FALSE;

    cinfo = get_info(c);

#ifdef DEBUG
    g_print("display_card %d %d,%d %s\n", cinfo-info, c->s, c->v,
	    hidden ? "hidden" : "not hidden");
#endif /* DEBUG */

    if (cinfo->stack != sn) {
	move_card(cinfo, sn, card_overlap*ypos);

	if (hidden) {
	    cinfo->state = CARD_HIDDEN;
	    set_card_hidden(cinfo);
	} else {
	    cinfo->state = CARD_VISIBLE;
	    set_card_visible(cinfo);
	}
	ret = TRUE;
    }

    if (cinfo->state == CARD_HIDDEN && !hidden) {
	cinfo->state = CARD_VISIBLE;
	set_card_visible(cinfo);

	ret = TRUE;
    }

    if (c->n == NULL && !hidden)
	show_card(cinfo);
    else
	show_card_small(cinfo);

    return ret;
}

static void
display_stack(stacknum_t sn)
{
    card_t *c;
    int ypos = 0;

#ifdef DEBUG
    g_print("display_stack %d\n", sn);
#endif /* DEBUG */

    for (c = layout[sn].hidden; c != nil; c = c->n) {
	display_card(c, TRUE, sn, ypos);
	ypos++;
    }

    for (c = layout[sn].faceup; c != nil; c = c->n)  {
	display_card(c, FALSE, sn, ypos);
	ypos++;
    }

    gtk_widget_queue_draw(GTK_WIDGET(stack[sn]));
}

void
display_layout(void)
{
    stacknum_t sn;

#ifdef DEBUG
    g_print("display_layout\n");
#endif /* DEBUG */

    for (sn = STACKS; sn > 0; --sn)
	display_stack(sn);
}

void
display_stack_less(stacknum_t sn)
{
    card_t *c;
    cardinfo_t *cinfo;

#ifdef DEBUG
    g_print("display_stack %d\n", sn);
#endif /* DEBUG */

    cinfo = NULL;
    for (c = layout[sn].faceup; c != nil; c = c->n) {
	cinfo = get_info(c);
	if (display_card(c, FALSE, sn, 0))
	    update_screen_delay();
    }
}

void
display_stack_more(stacknum_t sn)
{
    card_t *c;
    cardinfo_t *cinfo;
    int ypos = 0;

#ifdef DEBUG
    g_print("display_stack %d\n", sn);
#endif /* DEBUG */

    c = layout[sn].hidden;
    while (c != nil) {
	c = c->n;
	ypos++;
    }

    for (c = layout[sn].faceup; c != nil; c = c->n) {
	cinfo = get_info(c);
	if (display_card(c, FALSE, sn, ypos)) {
	    update_screen_delay();
	}
	ypos++;
    }

#if 0
    gtk_widget_queue_draw(GTK_WIDGET(stack[sn]));
#endif
}

void
display_aces(void)
{
    suit_t s;
    card_t c;
    cardinfo_t *cinfo;

    static int ace_ypos[4];
    static int map[4];

#ifdef DEBUG
    g_print("display_layout\n");
#endif /* DEBUG */

    if (ace_ypos[0] == 0) {
	int spacing;

	spacing = layout_height/4 - card_height - 1;
	ace_ypos[0] = spacing/2;
	ace_ypos[1] = ace_ypos[0] + layout_height/4;
	ace_ypos[2] = ace_ypos[1] + layout_height/4;
	ace_ypos[3] = ace_ypos[2] + layout_height/4;
    }
    if (naces == 0)
	map[0] = map[1] = map[2] = map[3] = -1;

    for (s = hearts; s <= clubs; s++) {
	c.s = s;
	c.v = aces[s];
	c.n = nil;

	if (c.v > none) {
	    if (map[s] == -1)
		map[s] = naces++;

	    cinfo = get_info(&c);
	    if (cinfo->stack != 8) {
		move_card(cinfo, 8, ace_ypos[map[s]]);
		set_card_visible(cinfo);
		show_card(cinfo);
		cinfo->state = CARD_VISIBLE;
	    }
	}
    }
#if 0
    gtk_widget_queue_draw(GTK_WIDGET(stack[8]));
#endif
}

void
display_deck(void)
{
    cardinfo_t *cinfo;

#ifdef DEBUG
    g_print("display_deck\n");
#endif /* DEBUG */

    if (reveal_mode) {
	printf("reveal mode ikke lavet endnu.\n");
    } else {
	card_t *card;

	if (top == nil)
	    gtk_widget_hide(GTK_WIDGET(back.w));
	else
	    gtk_widget_show(GTK_WIDGET(back.w));

	for (card = top; card != nil; card = card->n) {
	    cinfo = get_info(card);
	    hide_card(cinfo);
	    cinfo->state = CARD_UNSEEN;
	}

	if (deck != nil) {
	    cinfo = get_info(deck);
	    show_card(cinfo);
	    set_card_visible(cinfo);
	    cinfo->state = CARD_VISIBLE;

#ifdef DEBUG
	    g_print("display_deck top %d %d/%d\n", 
		    cinfo-info, deck->s, deck->v);
#endif /* DEBUG */

	    for (card = deck->n; card != nil; card = card->n) {
		cinfo = get_info(card);
		hide_card(cinfo);
		cinfo->state = CARD_UNSEEN;
	    }
	}
    }
#if 0
    gtk_widget_queue_draw(GTK_WIDGET(stack[0]));
#endif
}

gboolean
click_on_back(guint y)
{
    gint relative = y - (layout_height - 2*card_height);
    return (0 <= relative) && (relative < card_height);
}


void
display_auto_mode(gboolean auto_mode)
{
    gtk_widget_set_sensitive(button_new, !auto_mode);
    gtk_widget_set_sensitive(button_same, !auto_mode);
}

static display_status_t status = DISPLAY_NONE;

void
display_help(gboolean help, gboolean can_move)
{
    display_status_t nstatus;

    switch (status) {
    case DISPLAY_NONE:
    case DISPLAY_CAN_MOVE:
    case DISPLAY_NO_MOVE:
	if (help) {
	    nstatus = (can_move ? DISPLAY_CAN_MOVE : DISPLAY_NO_MOVE);

	    if (nstatus != status) {
		status = nstatus;
		gtk_pixmap_set(feedback, 
			       feedback_images[status].face, 
			       feedback_images[status].mask);
		gtk_widget_show(GTK_WIDGET(feedback));
	    }
	} else
	    gtk_widget_hide(GTK_WIDGET(feedback));
	break;

    default:
	break;
    }
}


void
display_status(display_status_t s)
{
    status = s;

    if (status == DISPLAY_NONE) {
	gtk_widget_hide(GTK_WIDGET(feedback));
    } else {
	gtk_pixmap_set(feedback, 
		       feedback_images[status].face, 
		       feedback_images[status].mask);
	gtk_widget_show(GTK_WIDGET(feedback));
    }
}

