/* GTKabale - Solitaire game for lazy people

   Copyright (C) 1999 Ren Seindal (rene@seindal.dk)

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA. 
 */

#include <glib.h>

#include "game.h"
#include "data.h"
#include "display.h"

#include <unistd.h>

static boolean help_mode;
static boolean auto_mode;
static boolean finished;

static guint delay;


typedef enum {
    move_flip_deck,
    move_deck_stack,
    move_stack_stack,
    move_stack_ace,
    move_deck_ace
} move_t;



void
set_help_mode(gboolean mode)
{
    help_mode = mode;
    display_help(help_mode, find_move(NULL, NULL));
}

void
start_game(gboolean same)
{
    initdata(same);

    pop_deck();

    display_clear();
    display_layout();
    display_aces();
    display_deck();
    display_info();

    display_status(DISPLAY_NONE);
    display_help(help_mode, find_move(NULL, NULL));

    finished = FALSE;
    auto_mode = FALSE;

}


static move_t
mapmove(stackindex_t src, stackindex_t dest)
{
    if (src == 0)
	if (dest == 0 || dest == 8)
	    return move_deck_ace;
	else
	    return move_deck_stack;
    else
	if (dest == 0 || dest == 8)
	    return move_stack_ace;
	else
	    return move_stack_stack;
}

static void
handle_move(move_t move, stackindex_t move_from, stackindex_t move_to)
{
    switch (move) {

    case move_flip_deck:
	if (!pop_deck()) {
	    --turnsleft;
	    top = revert_deck(deck);
	    deck = nil;
	    pop_deck();
	}
	if (turnsleft > 0)
	    display_deck();
	break;

    case move_stack_stack:
	if (move_stacks(move_from, move_to)) {
	    display_stack_more(move_to);
	    display_stack_less(move_from);
	}
	break;

    case move_stack_ace:
	if (move_stack_to_ace(move_from)) {
	    display_aces();
	    display_stack_less(move_from);
	}
	break;

    case move_deck_ace:
	if (move_deck_to_ace()) {
	    display_aces();
	    display_deck();
	}
	break;

    case move_deck_stack:
	if (move_deck_to_stack(move_to)) {
	    display_stack_more(move_to);
	    display_deck();
	}
	break;
    }
}



static void
do_auto_finish()
{
    stackindex_t move_from; 
    stackindex_t move_to;
    boolean has_move;
    move_t move;

    auto_mode = TRUE;
    delay = 100000;

    display_auto_mode(auto_mode);

    has_move = TRUE;
    while (auto_mode && has_move) {
	display_status(DISPLAY_NONE);

#ifdef DEBUG
	g_print("do_auto_finish_move\n");
#endif /* DEBUG */

	has_move = find_move(&move_from, &move_to);
	if (help_mode)
	    display_help(help_mode, has_move);

	if (has_move) {
	    move = mapmove(move_from, move_to);
	    handle_move(move, move_from, move_to);

	    display_info();
	    display_help(help_mode, has_move);
	} else {
	    times_played++;
	    if (succes())
		times_won++;

	    display_info();
	    display_status(succes() ? DISPLAY_WON : DISPLAY_LOST);
	}

	update_screen();
	usleep(delay);
    }

    auto_mode = FALSE;
    display_auto_mode(auto_mode);
    update_screen();
}

void
make_one_move(gint stack)
{
    static boolean deferred_move = FALSE;
    static stackindex_t deferred_from;

    stackindex_t move_to;
    move_t move;
    int nmoves;

    if (auto_mode)
	return;

    if (succes() || failure()) {
	start_game(FALSE);
	return;
    }

#ifdef DEBUG
    g_print("make_one_move %d (def %d)\n", stack, 
	    deferred_move ? deferred_from : -1);
#endif /* DEBUG */

    display_status(DISPLAY_NONE);

    if (stack < 0) {
	handle_move(move_flip_deck, 0, 0);
	deferred_from = FALSE;

	if (failure()) {
	    do_auto_finish();
	}
    } else if (deferred_move) {
	move = mapmove(deferred_from, stack);
	handle_move(move, deferred_from, stack);
	deferred_move = FALSE;
    } else {
	nmoves = find_unique_move(stack, &move_to);

	if (nmoves == 1) {
	    move = mapmove(stack, move_to);
	    handle_move(move, stack, move_to);
	} else if (nmoves > 1){
	    deferred_move = TRUE;
	    deferred_from = stack;
	    display_status(DISPLAY_AMBIGUOUS_MOVE);

#ifdef DEBUG
	    g_print("make_one_move deferring %d\n", stack);
#endif /* DEBUG */

	} else {
	    display_status(DISPLAY_BAD_MOVE);
	}
    }

    if (!auto_mode && (succes() || failure())) {
	times_played++;
	if (succes())
	    times_won++;
	display_info();

	display_status(succes() ? DISPLAY_WON : DISPLAY_LOST);
    }

    display_info();
    display_help(help_mode, help_mode ? find_move(NULL, NULL) : FALSE);
}



void
finish_game(void)
{
    stackindex_t move_from; 
    stackindex_t move_to;
    boolean has_move;
    move_t move;

#ifdef DEBUG
    g_print("finish_game\n");
#endif /* DEBUG */

    if (auto_mode) {
	delay = 0;
	return;
    }

    auto_mode = TRUE;
    display_auto_mode(auto_mode);
    delay = 100000;

    while (auto_mode && !succes() && !failure()) {
	display_status(DISPLAY_NONE);

#ifdef DEBUG
	g_print("finish_game move\n");
#endif /* DEBUG */

	has_move = find_move(&move_from, &move_to);
	if (help_mode)
	    display_help(help_mode, has_move);

	if (has_move)
	    move = mapmove(move_from, move_to);
	else
	    move = move_flip_deck;
	handle_move(move, move_from, move_to);

	display_info();

	if (succes() || failure()) {
	    times_played++;
	    if (succes())
		times_won++;

	    display_status(succes() ? DISPLAY_WON : DISPLAY_LOST);
	} else
	    display_help(help_mode, has_move);

	update_screen();
	usleep(delay);
    }

    auto_mode = FALSE;
    display_auto_mode(auto_mode);
    update_screen();
}
